
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "blocks.h"
#include "../Share/global.h"


static I_type *L;

static void get_block();
static void set_depth();


void
get_instance( problem )
    int		problem;
{
    int		i;
    char	name[30];
    FILE	*fp;

    if( problem < 10 ) sprintf( name, "./Problems/p0%1d", problem );
    else	       sprintf( name, "./Problems/p%2d",  problem );

    if( (fp = fopen( name, "r" )) == NULL ) {
	printf( "Could not open file %s\n", name );
	exit( 0 );
    }

    L = (I_type *)new_calloc( 1, sizeof(I_type) );

    /*
     *  Get and set parameters.
     */
    L->problem = problem;
    fscanf( fp, "Parameters:\n" );
    fscanf( fp, "\tname: %s\n", &name[0] );
    fscanf( fp, "\tblocks: %d\n", &(L->blocks) );

    /*
     *  Allocate space to store the initial and goal states.
     */
    L->init_above = new_calloc( L->blocks+1, sizeof(int) );
    L->init_below = new_calloc( L->blocks+1, sizeof(int) );
    L->goal_above = new_calloc( L->blocks+1, sizeof(int) );
    L->goal_below = new_calloc( L->blocks+1, sizeof(int) );
    L->init_depth = new_calloc( L->blocks+1, sizeof(int) );
    L->goal_depth = new_calloc( L->blocks+1, sizeof(int) );

    /*
     *  Get initial state.
     */
    fscanf( fp, "Initial:\n" );
    for( i = 1; i <= L->blocks; i++ ) {
	get_block( fp, L->init_above, L->init_below );
    }
    set_depth( L->init_depth, L->init_above );

    /*
     *  Get goal state.
     */
    fscanf( fp, "Goal:\n" );
    for( i = 1; i <= L->blocks; i++ ) {
	get_block( fp, L->goal_above, L->goal_below );
    }
    set_depth( L->goal_depth, L->goal_above );

    fclose( fp );
}

void
print_instance()
{
    printf( "problem: %2d  (%d blocks)\n", L->problem, L->blocks );
}

void
dump_instance()
{
    int		i;

    printf( "Initial:\n" );
    for( i = 1; i <= L->blocks; i++ ) {
	if( L->init_above[i] == 0 )
	    printf( "\tABOVE, %d, CLEAR\n", i );
	else
	    printf( "\tABOVE, %d, BLOCK, %d\n", i, L->init_above[i] );
	if( L->init_below[i] == 0 )
	    printf( "\tBELOW, %d, TABLE\n", i );
	else
	    printf( "\tBELOW, %d, BLOCK, %d\n", i, L->init_below[i] );
    }
    for( i = 1; i <= L->blocks; i++ ) {
	printf( "\tBlock %d, depth %d\n", i, L->init_depth[i] );
    }
    printf( "Goal:\n" );
    for( i = 1; i <= L->blocks; i++ ) {
	if( L->goal_above[i] == 0 )
	    printf( "\tABOVE, %d, CLEAR\n", i );
	else
	    printf( "\tABOVE, %d, BLOCK, %d\n", i, L->goal_above[i] );
	if( L->goal_below[i] == 0 )
	    printf( "\tBELOW, %d, TABLE\n", i );
	else
	    printf( "\tBELOW, %d, BLOCK, %d\n", i, L->goal_below[i] );
    }
    for( i = 1; i <= L->blocks; i++ ) {
	printf( "\tBlock %d, depth %d\n", i, L->goal_depth[i] );
    }
}


int
same_base( block )
    int		block;
{
    while( L->init_below[block] == L->goal_below[block] &&
	   L->init_below[block] != 0 ) {
	block = L->init_below[block];
    }

    return( L->init_below[block] == L->goal_below[block] );
}

/*
 *  Is block B below block A in a destination tower?
 */
int
precedence( b, a )
    int		b, a;
{
    while( L->goal_above[b] != a &&
	   L->goal_above[b] != 0 ) {
	b = L->goal_above[b];
    }

    return( L->goal_above[b] == a );
}


int get_num_blocks()		{ return( L->blocks ); }
int get_init_above( int a )	{ return( L->init_above[a] ); }
int get_init_below( int a )	{ return( L->init_below[a] ); }
int get_goal_above( int a )	{ return( L->goal_above[a] ); }
int get_goal_below( int a )	{ return( L->goal_below[a] ); }

int get_init_depth( int a )	{ return( L->init_depth[a] ); }
int get_goal_depth( int a )	{ return( L->goal_depth[a] ); }

int clear_in_init(    int a )	{ return( L->init_above[a] == 0 ); }
int clear_in_goal(    int a )	{ return( L->goal_above[a] == 0 ); }
int on_table_in_init( int a )	{ return( L->init_below[a] == 0 ); }
int on_table_in_goal( int a )	{ return( L->goal_below[a] == 0 ); }

int     get_lower_bound()	{ return( 2 ); }
int     get_upper_bound()	{ return( 4 * L->blocks ); }


static void
set_depth( depth, above )
    int		depth[], above[];
{
    int		i, b, count;

    for( i = 1; i <= L->blocks; i++ ) {
	count = 1;
	b = above[i];
	while( b != 0 ) {
	    count++;
	    b = above[b];
	}
	depth[i] = count;
    }
}

static void
get_block( fp, above, below )
    FILE	*fp;
    int		above[], below[];
{
    int		block1, block2;
    char	line[80];

    fgets( line, 80, fp ); 
    if( sscanf(line, "\tABOVE, %d, BLOCK, %d\n", &block1, &block2) == 2 ) {
	above[block1] = block2;
    }
    else
    if( sscanf(line, "\tABOVE, %d, CLEAR\n", &block1) == 1 ) {
	above[block1] = 0;
    }
    else {
	printf( "error: %s\n", line );
	exit( 6 );
    }

    fgets( line, 80, fp ); 
    if( sscanf(line, "\tBELOW, %d, BLOCK, %d\n", &block1, &block2) == 2 ) {
	below[block1] = block2;
    }
    else
    if( sscanf(line, "\tBELOW, %d, TABLE\n", &block1) == 1 ) {
	below[block1] = 0;
    }
    else {
	printf( "error: %s\n", line );
	exit( 6 );
    }
}

