
/*
 *  CPlan solves planning problems formulated as constraint satisfaction
 *  problems.
 *
 *  Copyright (C) 1999  Peter van Beek and Xinguang Chen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.  See the GNU General Public License
 *  for more details (see the file called Copying or contact the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).
 */

#include "blocks.h"
#include "../Share/global.h"
#include "../Share/csp.h"

static int	check_distance_above_constraint();
static int	check_distance_below_constraint();
static int	check_distance2_above_constraint();
static int	check_distance2_below_constraint();
static int	check_same_block_constraint();
static int	check_diff_block_constraint();
static int	check_action_above_constraint();
static int	check_action_below_constraint();
static int	check_useless2_action_constraint();
static int	check_useless3_action_constraint();
static int	check_useless4_action_constraint();
static int	check_symmetry_constraint();
static int	check_useful_action_constraint();

static int	no_change();

/*
 *  Check the given values against the constraint.
 */
int
check_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    checks += C->arity[r];

    switch( C->relation[r] ) {

    case DISTANCE_ABOVE_CONSTRAINT:
	return( check_distance_above_constraint( C, r, solution ) );

    case DISTANCE_BELOW_CONSTRAINT:
	return( check_distance_below_constraint( C, r, solution ) );

    case DISTANCE2_ABOVE_CONSTRAINT:
	return( check_distance2_above_constraint( C, r, solution ) );

    case DISTANCE2_BELOW_CONSTRAINT:
	return( check_distance2_below_constraint( C, r, solution ) );

    case SAME_BLOCK_CONSTRAINT:
	return( check_same_block_constraint( C, r, solution ) );

    case DIFF_BLOCK_CONSTRAINT:
	return( check_diff_block_constraint( C, r, solution ) );

    case ACTION_ABOVE_CONSTRAINT:
	return( check_action_above_constraint( C, r, solution ) );

    case ACTION_BELOW_CONSTRAINT:
	return( check_action_below_constraint( C, r, solution ) );

    case USELESS2_ACTIONS_CONSTRAINT:
	return( check_useless2_action_constraint( C, r, solution ) );

    case USELESS3_ACTIONS_CONSTRAINT:
	return( check_useless3_action_constraint( C, r, solution ) );

    case USELESS4_ACTIONS_CONSTRAINT:
	return( check_useless4_action_constraint( C, r, solution ) );

    case SYMMETRY_CONSTRAINT:
	return( check_symmetry_constraint( C, r, solution ) );

    case USEFUL_ACTION_CONSTRAINT:
	return( check_useful_action_constraint( C, r, solution ) );

    default:
	printf("Error in check_constraint.\n");
	exit(6);
    }
}

static int
same_attributes( attr1, attr2 )
    VALA	 attr1, attr2;
{
    return( attr1.type   == attr2.type &&
	    attr1.number == attr2.number );
}

static int
check_action_above_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int		i, j, k, l;
    int		a, b, c, d;

    i = C->scheme[r][0]; /* above_a(t)	 */
    j = C->scheme[r][1]; /* arm(t)	 */
    k = C->scheme[r][2]; /* above_a(t+1) */
    l = C->scheme[r][3]; /* arm(t+1)	 */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    if( C->value_attribute[i][a].type == CLEAR &&
        C->value_attribute[k][c].type == HELD &&
        !(C->value_attribute[j][b].type == EMPTY &&
          C->value_attribute[l][d].type == BLOCK &&
          C->value_attribute[l][d].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == CLEAR &&
        C->value_attribute[k][c].type == BLOCK &&
        !(C->value_attribute[j][b].type == BLOCK &&
          C->value_attribute[l][d].type == EMPTY &&
          C->value_attribute[j][b].number ==
    	    C->value_attribute[k][c].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == HELD &&
        C->value_attribute[k][c].type == CLEAR &&
        !(C->value_attribute[j][b].type == BLOCK &&
          C->value_attribute[l][d].type == EMPTY &&
          C->value_attribute[j][b].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == BLOCK &&
        C->value_attribute[k][c].type == CLEAR &&
        !(C->value_attribute[j][b].type == EMPTY &&
          C->value_attribute[l][d].type == BLOCK &&
          C->value_attribute[l][d].number ==
    	    C->value_attribute[i][a].number) ) {
	return( 0 );
    }
    /*
     *  Because the arm is a unary resource ...
     */
    else
    if(  C->value_attribute[j][b].type == EMPTY &&
         C->value_attribute[l][d].type == BLOCK &&
         C->value_attribute[l][d].number !=
    	    C->variable_attribute[i].number &&
	!(C->value_attribute[i][a].type == BLOCK &&
	  C->value_attribute[k][c].type == CLEAR &&
	  C->value_attribute[i][a].number ==
             C->value_attribute[l][d].number) &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[k][c] ) ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[j][b].type == BLOCK &&
         C->value_attribute[l][d].type == EMPTY &&
         C->value_attribute[j][b].number !=
    	    C->variable_attribute[i].number &&
	!(C->value_attribute[i][a].type == CLEAR &&
	  C->value_attribute[k][c].type == BLOCK &&
	  C->value_attribute[k][c].number ==
             C->value_attribute[j][b].number) &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[k][c] ) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

static int
check_action_below_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int		i, j, k, l;
    int		a, b, c, d;

    i = C->scheme[r][0]; /* below_a(t)	 */
    j = C->scheme[r][1]; /* arm(t)	 */
    k = C->scheme[r][2]; /* below_a(t+1) */
    l = C->scheme[r][3]; /* arm(t+1)	 */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    if( C->value_attribute[i][a].type == HELD &&
        C->value_attribute[k][c].type == TABLE &&
        !(C->value_attribute[j][b].type == BLOCK &&
          C->value_attribute[l][d].type == EMPTY &&
          C->value_attribute[j][b].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == TABLE &&
        C->value_attribute[k][c].type == HELD &&
        !(C->value_attribute[j][b].type == EMPTY &&
          C->value_attribute[l][d].type == BLOCK &&
          C->value_attribute[l][d].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == HELD &&
        C->value_attribute[k][c].type == BLOCK &&
        !(C->value_attribute[j][b].type == BLOCK &&
          C->value_attribute[l][d].type == EMPTY &&
          C->value_attribute[j][b].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == BLOCK &&
        C->value_attribute[k][c].type == HELD &&
        !(C->value_attribute[j][b].type == EMPTY &&
          C->value_attribute[l][d].type == BLOCK &&
          C->value_attribute[l][d].number ==
    	    C->variable_attribute[i].number) ) {
	return( 0 );
    }
    /*
     *  Because the arm is a unary resource ...
     */
    else
    if(  C->value_attribute[j][b].type == EMPTY &&
         C->value_attribute[l][d].type == BLOCK &&
         C->value_attribute[l][d].number !=
    	    C->variable_attribute[i].number &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[k][c] ) ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[j][b].type == BLOCK &&
         C->value_attribute[l][d].type == EMPTY &&
         C->value_attribute[j][b].number !=
    	    C->variable_attribute[i].number &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[k][c] ) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a block variable.
 */
static int
distance_above( a, a_new )
    VALA	 a, a_new;
{
    if( a.type == BLOCK && a_new.type == BLOCK ) {
	if( a.number == a_new.number ) return( 0 );
	else			       return( 4 );
    }
    else if( a.type == BLOCK && a_new.type == CLEAR ) return( 1 );
    else if( a.type == BLOCK && a_new.type == HELD  ) return( 3 );
    else if( a.type == CLEAR && a_new.type == BLOCK ) return( 1 );
    else if( a.type == CLEAR && a_new.type == CLEAR ) return( 0 );
    else if( a.type == CLEAR && a_new.type == HELD  ) return( 1 );
    else if( a.type == HELD  && a_new.type == BLOCK ) return( 3 );
    else if( a.type == HELD  && a_new.type == CLEAR ) return( 1 );
    else if( a.type == HELD  && a_new.type == HELD  ) return( 4 );
    else {
	printf("Error in distance_above.\n");
	exit(6);
    }
}

static int
check_distance_above_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* block(t1) */
    j = C->scheme[r][1];  /* block(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance_above( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of a block variable.
 */
static int
distance_below( a, a_new )
    VALA	 a, a_new;
{
    if( a.type == BLOCK && a_new.type == BLOCK ) {
	if( a.number == a_new.number ) return( 0 );
	else			       return( 2 );
    }
    else if( a.type == BLOCK && a_new.type == TABLE ) return( 2 );
    else if( a.type == BLOCK && a_new.type == HELD  ) return( 1 );
    else if( a.type == TABLE && a_new.type == BLOCK ) return( 2 );
    else if( a.type == TABLE && a_new.type == TABLE ) return( 0 );
    else if( a.type == TABLE && a_new.type == HELD  ) return( 1 );
    else if( a.type == HELD  && a_new.type == BLOCK ) return( 1 );
    else if( a.type == HELD  && a_new.type == TABLE ) return( 1 );
    else if( a.type == HELD  && a_new.type == HELD  ) return( 4 );
    else {
	printf("Error in distance_below.\n");
	exit(6);
    }
}

static int
check_distance_below_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* block(t1) */
    j = C->scheme[r][1];  /* block(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance_below( C->value_attribute[i][a], C->value_attribute[j][b] )
    );
}

static int
no_change( attr1, attr2 )
    VALA	 attr1, attr2;
{
    return( attr1.type   == attr2.type &&
	    attr1.number == attr2.number &&
	    attr1.type   != HELD );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of two block variables.
 */
static int
distance2_above( a, a_new,
		 b, b_new )
    VALA	a, a_new,
	        b, b_new;
{
    if( no_change( a, a_new ) ||
        no_change( b, b_new ) ) {
	return( 0 );
    }
    else
    if( a.type == BLOCK && a_new.type == BLOCK &&
        b.type == BLOCK && b_new.type == BLOCK ) {
	if( a.number != b_new.number &&
	    b.number != a_new.number )	return( 8 );
	else				return( 6 );
    }
    else
    if( a.type == BLOCK &&
        b.type == BLOCK ) {
	if( a_new.type == HELD &&
	    b_new.type == BLOCK ) {
	    if( a.number != b_new.number ) return( 7 );
	    else			   return( 5 );
	}
	else
	if( a_new.type == BLOCK &&
	    b_new.type == HELD ) {
	    if( b.number != a_new.number ) return( 7 );
	    else			   return( 5 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == BLOCK ) {
	    if( a.number != b_new.number ) return( 5 );
	    else			   return( 4 );
	}
	else
	if( a_new.type == BLOCK &&
	    b_new.type == CLEAR ) {
	    if( b.number != a_new.number ) return( 5 );
	    else			   return( 4 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == HELD ) {
	    return( 3 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == CLEAR ) {
	    return( 3 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == CLEAR ) {
	    return( 3 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == BLOCK &&
        b_new.type == BLOCK ) {
	if( a.type == HELD &&
	    b.type == BLOCK ) {
	    if( a_new.number != b.number ) return( 7 );
	    else			   return( 5 );
	}
	else
	if( a.type == BLOCK &&
	    b.type == HELD ) {
	    if( b_new.number != a.number ) return( 7 );
	    else			   return( 5 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == BLOCK ) {
	    if( a_new.number != b.number ) return( 5 );
	    else			   return( 4 );
	}
	else
	if( a.type == BLOCK &&
	    b.type == CLEAR ) {
	    if( b_new.number != a.number ) return( 5 );
	    else			   return( 4 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == HELD ) {
	    return( 3 );
	}
	else
	if( a.type == HELD &&
	    b.type == CLEAR ) {
	    return( 3 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == CLEAR ) {
	    return( 3 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a.type == BLOCK &&
        b.type == CLEAR ) {
	if( a_new.type == BLOCK &&
	    b_new.type == HELD ) {
	    return( 5 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == BLOCK ) {
	    if( a.number != b_new.number ) return( 4 );
	    else			   return( 3 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == BLOCK ) {
	    return( 2 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == HELD ) {
	    return( 1 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == BLOCK &&
        b_new.type == CLEAR ) {
	if( a.type == BLOCK &&
	    b.type == HELD ) {
	    return( 5 );
	}
	else
	if( a.type == HELD &&
	    b.type == BLOCK ) {
	    if( a_new.number != b.number ) return( 4 );
	    else			   return( 3 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == BLOCK ) {
	    return( 2 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == HELD ) {
	    return( 1 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a.type == CLEAR &&
        b.type == BLOCK ) {
	if( a_new.type == BLOCK &&
	    b_new.type == HELD ) {
	    if( b.number != a_new.number ) return( 4 );
	    else			   return( 3 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == BLOCK ) {
	    return( 5 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == CLEAR ) {
	    return( 1 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == CLEAR &&
        b_new.type == BLOCK ) {
	if( a.type == BLOCK &&
	    b.type == HELD ) {
	    if( b_new.number != a.number ) return( 4 );
	    else			   return( 3 );
	}
	else
	if( a.type == HELD &&
	    b.type == BLOCK ) {
	    return( 5 );
	}
	else
	if( a.type == HELD &&
	    b.type == CLEAR ) {
	    return( 1 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a.type == HELD &&
        b.type == BLOCK ) {
	if( a_new.type == BLOCK &&
	    b_new.type == HELD ) {
	    if( b.number != a_new.number ) return( 6 );
	    else			   return( 4 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == CLEAR ) {
	    return( 2 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == HELD ) {
	    return( 4 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == BLOCK ) {
	    return( 6 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == CLEAR ) {
	    return( 4 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == HELD &&
        b_new.type == BLOCK ) {
	if( a.type == BLOCK &&
	    b.type == HELD ) {
	    if( b_new.number != a.number ) return( 6 );
	    else			   return( 4 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == CLEAR ) {
	    return( 2 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == HELD ) {
	    return( 4 );
	}
	else
	if( a.type == HELD &&
	    b.type == BLOCK ) {
	    return( 6 );
	}
	else
	if( a.type == HELD &&
	    b.type == CLEAR ) {
	    return( 4 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a.type == BLOCK &&
        b.type == HELD ) {
	if( a_new.type == BLOCK &&
	    b_new.type == HELD ) {
	    return( 6 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == CLEAR ) {
	    return( 2 );
	}
	else
	if( a_new.type == CLEAR &&
	    b_new.type == HELD ) {
	    return( 4 );
	}
	else
	if( a_new.type == HELD &&
	    b_new.type == CLEAR ) {
	    return( 4 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == BLOCK &&
        b_new.type == HELD ) {
	if( a.type == BLOCK &&
	    b.type == HELD ) {
	    return( 6 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == CLEAR ) {
	    return( 2 );
	}
	else
	if( a.type == CLEAR &&
	    b.type == HELD ) {
	    return( 4 );
	}
	else
	if( a.type == HELD &&
	    b.type == CLEAR ) {
	    return( 4 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a.type == CLEAR &&
        b.type == HELD ) {
	if( a_new.type == HELD &&
	    b_new.type == CLEAR ) {
	    return( 2 );
	}
	else {
	    return( 0 );
	}
    }
    else
    if( a_new.type == CLEAR &&
        b_new.type == HELD ) {
	if( a.type == HELD &&
	    b.type == CLEAR ) {
	    return( 2 );
	}
	else {
	    return( 0 );
	}
    }
    else {
	return( 0 );
    }
}

/*
 *  Check 4-ary distance constraint.
 */
static int
check_distance2_above_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0];  /* block_a(t1) */
    j = C->scheme[r][1];  /* block_a(t2) */
    k = C->scheme[r][2];  /* block_b(t1) */
    l = C->scheme[r][3];  /* block_b(t2) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];

    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance2_above( C->value_attribute[i][a], C->value_attribute[j][b],
			 C->value_attribute[k][c], C->value_attribute[l][d] )
    );
}

/*
 *  Determine a lower bound on the number of time steps that would
 *  be needed to change the value of two block variables.
 */
static int
distance2_below( a, a_new,
		 b, b_new )
    VALA	a, a_new,
	        b, b_new;
{
    if( no_change( a, a_new ) ||
        no_change( b, b_new ) ) {
	return( 0 );
    }
    else
    if( a.type == BLOCK && a_new.type == BLOCK &&
        b.type == BLOCK && b_new.type == BLOCK ) {
	if( a.number == b_new.number &&
	    b.number == a_new.number )	return( 6 );
	else				return( 4 );
    }
    else
    if( a.type == BLOCK && a_new.type == TABLE &&
        b.type == BLOCK && b_new.type == TABLE ) {
	return( 4 );
    }
    else
    if( a.type == TABLE && a_new.type == BLOCK &&
        b.type == TABLE && b_new.type == BLOCK ) {
	return( 4 );
    }
    else
    if( a.type == BLOCK && a_new.type == HELD &&
        b.type == BLOCK && b_new.type == BLOCK ) {
	if( a.number == b_new.number )	return( 5 );
	else				return( 3 );
    }
    else
    if( a.type == BLOCK && a_new.type == BLOCK &&
        b.type == BLOCK && b_new.type == HELD ) {
	if( b.number == a_new.number )	return( 5 );
	else				return( 3 );
    }
    else {
	return( 0 );
    }
}

/*
 *  Check 4-ary distance constraint.
 */
static int
check_distance2_below_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k, l;
    int         a, b, c, d;

    i = C->scheme[r][0];  /* block_a(t1) */
    j = C->scheme[r][1];  /* block_a(t2) */
    k = C->scheme[r][2];  /* block_b(t1) */
    l = C->scheme[r][3];  /* block_b(t2) */

    a = solution[i];
    b = solution[j];
    c = solution[k];
    d = solution[l];


    /*
     *  actual distance >= lower bound on distance
     */
    return(
	(C->variable_attribute[j].state - C->variable_attribute[i].state) >=
	distance2_below( C->value_attribute[i][a], C->value_attribute[j][b],
			 C->value_attribute[k][c], C->value_attribute[l][d] )
    );
}

static int
check_same_block_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* a_above(t) */
    j = C->scheme[r][1];  /* a_below(t) */

    a = solution[i];
    b = solution[j];

    if( (C->value_attribute[i][a].type == HELD &&
         C->value_attribute[j][b].type != HELD) ||
        (C->value_attribute[j][b].type == HELD &&
         C->value_attribute[i][a].type != HELD) ) {
	return( 0 );
    }
    else
    if( C->value_attribute[i][a].type == BLOCK &&
        C->value_attribute[j][b].type == BLOCK &&
        C->value_attribute[i][a].number ==
            C->value_attribute[j][b].number ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

static int
check_diff_block_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;

    i = C->scheme[r][0];  /* a_above(t) */
    j = C->scheme[r][1];  /* b_below(t) */

    a = solution[i];
    b = solution[j];

    if(  C->value_attribute[i][a].type == BLOCK &&
         C->value_attribute[i][a].number ==
    	     C->variable_attribute[j].number &&
	!(C->value_attribute[j][b].type == BLOCK &&
          C->value_attribute[j][b].number ==
    	     C->variable_attribute[i].number) ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[j][b].type == BLOCK &&
         C->value_attribute[j][b].number ==
    	     C->variable_attribute[i].number &&
	!(C->value_attribute[i][a].type == BLOCK &&
          C->value_attribute[i][a].number ==
    	     C->variable_attribute[j].number) ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[i][a].type == CLEAR &&
         C->value_attribute[j][b].type == BLOCK &&
    	 C->variable_attribute[i].number ==
             C->value_attribute[j][b].number ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[i][a].type == BLOCK &&
         C->value_attribute[j][b].type == TABLE &&
    	 C->variable_attribute[j].number ==
             C->value_attribute[i][a].number ) {
	return( 0 );
    }

    else
    if(  C->value_attribute[i][a].type == HELD &&
         C->value_attribute[j][b].type == BLOCK &&
    	 C->variable_attribute[i].number ==
             C->value_attribute[j][b].number ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[i][a].type == BLOCK &&
         C->value_attribute[j][b].type == HELD &&
    	 C->variable_attribute[j].number ==
             C->value_attribute[i][a].number ) {
	return( 0 );
    }

    else
    if( C->value_attribute[i][a].type == HELD &&
        C->value_attribute[j][b].type == HELD ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Prevent useless actions constraint.
 */
static int
check_useless2_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j;
    int         a, b;
    int		dest_block;

    i = C->scheme[r][0]; /* block_a(t1) */
    j = C->scheme[r][1]; /* block_a(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];

    /* get destination block */
    if( C->variable_attribute[i].type == ABOVE )
	dest_block = get_goal_above( C->variable_attribute[i].number );
    else
	dest_block = get_goal_below( C->variable_attribute[i].number );

    /*
     *  Once a block is moved above or below its
     *  destination block, it stays there.
     */
    if(  C->value_attribute[i][a].type == BLOCK &&
         C->value_attribute[i][a].number == dest_block &&
	!same_attributes( C->value_attribute[i][a],
			  C->value_attribute[j][b] ) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *  Prevent useless actions constraint.
 */
static int
check_useless3_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k;
    int         a, b, c, changes;

    i = C->scheme[r][0]; /* block_a(t1) */
    j = C->scheme[r][1]; /* block_a(t2) */
    k = C->scheme[r][2]; /* block_a(t3), 0 <= t1 < t2 < t3 <= time */

    a = solution[i];
    b = solution[j];
    c = solution[k];

    if( C->value_attribute[j][b].type == BLOCK ) {
	changes = 0;
	if( !same_attributes( C->value_attribute[i][a],
			      C->value_attribute[j][b] ) ) changes++;
	if( !same_attributes( C->value_attribute[j][b],
			      C->value_attribute[k][c] ) ) changes++;
	if( changes > 1 )
	    return( 0 );
	else
	    return( 1 );
    }
    else {
	return( 1 );
    }
}



/*
 *  Prevent useless actions constraint.
 */
static int
check_useless4_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k;
    int         a, b, c;

    i = C->scheme[r][0]; /* below_a(t1) */
    j = C->scheme[r][1]; /* below_a(t1+1) */
    k = C->scheme[r][2]; /* below_a(t2), 0 <= t1 < t2 <= time */

    a = solution[i];
    b = solution[j];
    c = solution[k];

    if(  C->value_attribute[i][a].type == TABLE &&
         C->value_attribute[j][b].type == HELD &&
        (C->value_attribute[k][c].type == TABLE ||
         C->value_attribute[k][c].type == HELD) ) {
	return( 0 );
    }
    else
    if(  C->value_attribute[i][a].type == HELD &&
         C->value_attribute[j][b].type == HELD &&
        (C->value_attribute[k][c].type == TABLE ||
         C->value_attribute[k][c].type == HELD) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *
 */
static int
check_symmetry_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         h, i, j, k, l;
    int         a, b, c, d, e;

    h = C->scheme[r][0]; /* above_a(t0) */
    i = C->scheme[r][1]; /* above_b(t0) */
    j = C->scheme[r][2]; /* above_b(t0+1) */
    k = C->scheme[r][3]; /* below_a(t1) */
    l = C->scheme[r][4]; /* below_a(t1+1) */

    a = solution[h];
    b = solution[i];
    c = solution[j];
    d = solution[k];
    e = solution[l];

    if( C->value_attribute[h][a].type == CLEAR &&
        C->value_attribute[i][b].type == CLEAR &&
        C->value_attribute[j][c].type == HELD  &&
        C->value_attribute[k][d].type == HELD  &&
        C->value_attribute[l][e].type == TABLE ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

/*
 *
 */
static int
check_useful_action_constraint( C, r, solution )
    CSP_type    *C;
    int         r;
    int         *solution;
{
    int         i, j, k;
    int         a, b, c;
    int		number_a, type_b, number_b;

    i = C->scheme[r][0]; /* below_a(t)   */
    j = C->scheme[r][1]; /* below_a(t+1) */
    k = C->scheme[r][2]; /* below_b(t+1) */

    a = solution[i];
    b = solution[j];
    c = solution[k];

    /* get destination blocks */
    number_a = get_goal_below( C->variable_attribute[j].number );
    number_b = get_goal_below( C->variable_attribute[k].number );
    if( on_table_in_goal( C->variable_attribute[k].number ) ) {
	type_b = TABLE;
    }
    else {
	type_b = BLOCK;
    }

    if(  C->value_attribute[i][a].type == HELD &&
         C->value_attribute[j][b].type == BLOCK &&
       !(C->value_attribute[j][b].number == number_a &&
         C->value_attribute[k][c].type == type_b &&
         C->value_attribute[k][c].number == number_b) ) {
	return( 0 );
    }
    else {
	return( 1 );
    }
}

